// File:           bllib.cpp - code module for BLlib
//                 -*- C++ -*-
//
// Package:        Library
//
// Author:         Simone Zanella
//
// E-Mail:         szanella@dsi.unive.it
//
// Address:        via Gramsci, 72/22
//                 30035 Mirano (VE)
//                 ITALY
//
// Copyright (C) 1997 Simone Zanella Productions
//
// This program is free software; you can redistribute
// it and/or modify it under the terms of GNU General Public License
// as published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implicit warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// History:
//
// $Log:	bllib.cpp,v $
// Revision 1.13  97/03/28  12:00:00  12:00:00  szanella (Zanella Simone)
// first stand-alone public release
//


//******************************************************************
//************************ INCLUDE SECTION *************************
//******************************************************************

#include "bllib.h"

#include <math.h>
#include <float.h>
#include <limits.h>


//******************************************************************
//*********************** CONSTANT SECTION *************************
//******************************************************************


enum TypeSignatures { _CHAR,   _INT16,  _INT32,  _INT64,
                      _UCHAR,  _UINT16, _UINT32, _UINT64,
                      _FLOAT,  _DOUBLE,
                      _CHARPT, _CHARPS, _CHARPL, _CHARPH,
                      _RDATAT, _RDATAS, _RDATAL, _RDATAH };

enum TypeOfIntegers { _BL_CHAR,  _BL_SHORT,  _BL_INT,  _BL_LONG,
                      _BL_UCHAR, _BL_USHORT, _BL_UINT, _BL_ULONG };

#define _BL_DOUBLE_SIZE 8
#define _BL_FLOAT_SIZE  4
#define _BL_CHAR_SIZE   1
#define _BL_SIGN_SIZE   1


//******************************************************************
//******************* PUBLIC VARIABLES SECTION *********************
//******************************************************************

// Public variables: these variables are used and automatically
// updated by the functions defined hereafter; the user must NOT modify
// them or rely on them to store values between function calls.


int BL_convResult = BL_CONV_OK;
// Result of item extraction or last conversion (BL_ToIntxx / BL_ToUintxx)


BL_Int16 BL_i16;              // Static data overwritten
BL_Uint16 BL_ui16;            // by the BL_ToIntxx & BL_ToUintxx
BL_Int32 BL_i32;              // functions.
BL_Uint32 BL_ui32;
BL_Int64 BL_i64;
BL_Uint64 BL_ui64;
BL_RawData BL_raw;            // Static data overwritten by BL_ToRawData


//******************************************************************
//************************ COMMON FUNCTIONS ************************
//******************************************************************

template <class T> T
BL_min(T x, T y)
{
  // Template for performing minimum evaluation

  return (x < y) ? x : y;
}


void
ByteRev(char *cpLENumber, int iSize)
{
  //
  // Function to invert a byte string in one byte blocks.
  //
  // e.g. 1234 -> 4321
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: none
  //
  // Parameters:   char *  cpPointer (pointer to memory object)
  //               int     iSize     (object' size)
  //
  // Returns:      void
  //

  register int iCursor = 0, iLimit;
  register char cTemp;

  iLimit = iSize/2; // Last byte to exchange
  for (; iCursor<iLimit; iCursor++)
  {
    cTemp = *(cpLENumber+iCursor);
    *(cpLENumber+iCursor) = *(cpLENumber+iSize-iCursor-1);
    *(cpLENumber+iSize-iCursor-1) = cTemp;
  }

  return;
}


void
ResizeMsg(char * (&cpMsgBuf), unsigned long ulNewLength,
          unsigned long ulOldLength, unsigned long &ulBufOffset)
{
  //
  // Function to resize the buffer allocated to a message.
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: BL_ChkNul;
  //
  // Parameters:   char * &cpMsgBuf     pointer to buffer
  //               ulong  ulNewLength   new length of buffer
  //
  // Returns:      void
  //

  BL_ChkNul(cpMsgBuf = (char *) realloc(cpMsgBuf, ulNewLength));
  if (ulOldLength==ulBufOffset)
    ++ulBufOffset;
  return;

}


int
CalcLen(unsigned char * ucaLength, unsigned long ulLength)
{

  //
  // Function to store a value in the minimum number of bytes.
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: none;
  //
  // Parameters:   uchar * ucaLength        array of max 4 char
  //               ulong   ulLength         length to store
  //
  // Returns:      number of bytes used in ucaLength
  //


  register int iUsedBytes = 0;

  do
  {
    ucaLength[iUsedBytes++] = ulLength % 256;
    ulLength /= 256;
  }
  while (ulLength);

  return iUsedBytes;

}


unsigned long
GiveLen(char * caLength, int iUsedBytes)
{

  //
  // Function to return the value of an unsigned long previously
  // stored with CalcLen.
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: none;
  //
  // Parameters:   char *  caLength         array of max 4 char
  //               int     iUsedBytes       number of bytes used
  //                                        to store the number
  //
  // Returns:      value stored
  //


  register unsigned long ulDivisor = 1, ulNumber = 0;

  for (int x=0; x<iUsedBytes; x++)
  {
    ulNumber += (ulDivisor * (unsigned char) caLength[x]);
    ulDivisor *= 256;
  }

  return ulNumber;

}


unsigned long
CalcItemSize(char * cpBuffer)
{

  //
  // Function to calculate the size of an item (number of bytes taken
  // by the signature + the size + item's data).
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: GiveLen;
  //
  // Parameters:   char * cpBuffer           pointer to start of item
  //
  // Returns:      number of bytes allocated to item
  //


  unsigned long ulItemSize = _BL_SIGN_SIZE;
  char cTypeSign;


  // Checks type signature & calculates item' size
  switch (cTypeSign = *(cpBuffer))
  {
    case _CHAR: case _UCHAR:
      ulItemSize += _BL_CHAR_SIZE;
      break;

    case _INT16: case _UINT16:
      ulItemSize += 2;
      break;

    case _INT32: case _UINT32:
      ulItemSize += 4;
      break;

    case _INT64: case _UINT64:
      ulItemSize += 8;
      break;

    case _FLOAT:
      ulItemSize += _BL_FLOAT_SIZE;
      break;

    case _DOUBLE:
      ulItemSize += _BL_DOUBLE_SIZE;
      break;

    case _CHARPT: case _CHARPS: case _CHARPL: case _CHARPH:
      // warning: short cut for value!
      ulItemSize += (cTypeSign - _CHARPT + 1);
      ulItemSize += GiveLen(cpBuffer + _BL_SIGN_SIZE, cTypeSign - _CHARPT + 1);
      break;

    case _RDATAT: case _RDATAS: case _RDATAL: case _RDATAH:
      // warning: short cut for value!
      ulItemSize += (cTypeSign - _RDATAT + 1);
      ulItemSize += GiveLen(cpBuffer + _BL_SIGN_SIZE, cTypeSign - _RDATAT + 1);
      break;
  }

  return ulItemSize;

}


int
SignatureToPreset(char cType)
{
  //
  // Function to transform a type signature into a preset type;
  // it remaps, for example, different _CHARPx signatures (tiny, small,
  // large and huge) to the same preset BL_CHARP type.
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: none;
  //
  // Parameters:   char cType                type signature
  //
  // Returns:      type value, or BL_NOT_VALID if item type is unknown
  //


  int iRetValue = BL_NOT_VALID;

  switch (cType)
  {
    case _CHAR:
      iRetValue = BL_CHAR;
      break;
    case _UCHAR:
      iRetValue = BL_UCHAR;
      break;
    case _INT16:
      iRetValue = BL_I16;
      break;
    case _UINT16:
      iRetValue = BL_UI16;
      break;
    case _INT32:
      iRetValue = BL_I32;
      break;
    case _UINT32:
      iRetValue = BL_UI32;
      break;
    case _INT64:
      iRetValue = BL_I64;
      break;
    case _UINT64:
      iRetValue = BL_UI64;
      break;
    case _FLOAT:
      iRetValue = BL_FLOAT;
      break;
    case _DOUBLE:
      iRetValue = BL_DOUBLE;
      break;
    case _CHARPT: case _CHARPS: case _CHARPL: case _CHARPH:
      iRetValue = BL_CHARP;
      break;
    case _RDATAT: case _RDATAS: case _RDATAL: case _RDATAH:
      iRetValue = BL_RAWDATA;
      break;
  }

  return iRetValue;

}


unsigned long
GiveSizeOf(int iPreset)
{

  //
  // Function to calculate the size of fixed-length types (int & float).
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: none;
  //
  // Parameters:   int iPreset              preset type
  //
  // Returns:      size in bytes of type data, 0 if unknown
  //


  unsigned long ulRetValue = 0;

  switch (iPreset)
  {
    case BL_CHAR: case BL_UCHAR:
      ulRetValue = _BL_CHAR_SIZE;
      break;
    case BL_I16: case BL_UI16:
      ulRetValue = 2;
      break;
    case BL_I32: case BL_UI32:
      ulRetValue = 4;
      break;
    case BL_I64: case BL_UI64:
      ulRetValue = 8;
      break;
    case BL_FLOAT:
      ulRetValue = _BL_FLOAT_SIZE;
      break;
    case BL_DOUBLE:
      ulRetValue = _BL_DOUBLE_SIZE;
      break;
  }

  return ulRetValue;

}


unsigned long
GiveSizeOf(char * cpBuffer)
{

  //
  // Function to calculate the size of variable-length types
  // (strings & rawdata).
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: GiveLen;
  //
  // Parameters:   char * cpBuffer          pointer to buffer
  //
  // Returns:      size in bytes of type data, 0 if unknown
  //


  unsigned long ulRetValue = 0;
  char cTypeSign;

  switch (cTypeSign = *cpBuffer)
  {
    case _CHARPT: case _CHARPS: case _CHARPL: case _CHARPH:
      // warning: short cut for value!
      ulRetValue = GiveLen(cpBuffer + _BL_SIGN_SIZE, cTypeSign - _CHARPT + 1);
      break;

    case _RDATAT: case _RDATAS: case _RDATAL: case _RDATAH:
      // warning: short cut for value!
      ulRetValue = GiveLen(cpBuffer + _BL_SIGN_SIZE, cTypeSign - _RDATAT + 1);
      break;
  }

  return ulRetValue;

}


char *
BL_ChkNul(char *cpMem)
{
  // Function to check result of allocation functions

  if (cpMem == NULL) // Checks result of allocation functions
  {
    cerr << "BLlib: not enough memory.";
    exit(1);
  }
  return cpMem;
}


BL_RawData
BL_ToRawData(char * cpBinData, unsigned long ulLength)
{

  //
  // Function to transform a pair <pointer to mem, mem length> into
  // a rawdata structure. It returns the public variable BL_raw
  // modified according to the values passed.
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: none;
  //
  // Parameters:   char *  cpBinData        pointer to a memory area
  //               ulong   ulLength         area length
  //
  // Returns:      BL_raw modified
  //


  BL_raw.pcData = cpBinData;
  BL_raw.ulDataLen = ulLength;

  return BL_raw;

}


//******************************************************************
//***************** INTEGER MANIPULATION FUNCTIONS *****************
//******************************************************************

// Functions used to convert an integer to the internal memory format.

int
PutUint (char * cpSource, char * cpDest, int iSourceSize, int iDestSize)
{
  //
  // Function to degrade an unsigned integer of a certain size to a known
  // size in big endian format used by BLlib.
  // The value is read from cpSource and stored in cpDest.
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: ByteRev;
  //
  // Parameters:   char * cpDest        pointer to dest. (already allocated)
  //               char * cpSource      pointer to source
  //               int    iSourceSize   source size (bytes)
  //               int    iDestSize     destination size (bytes)
  //
  // Returns:      BL_CONV_OK     if transformation succesfull
  //               BL_CANT_SHRINK if transformation failed
  //
  // Note:
  //               cpDest is set to all zeroes if degrading
  //               was not possible.
  //


  register int iCounter;
  register char ok = 1;

  memset(cpDest, 0, iDestSize);

  #ifdef BL_LITTLE_END

   // Checks if variable fits in the model
   if (iSourceSize>iDestSize)
     for (iCounter = iDestSize; iCounter<iSourceSize; iCounter++)
      if (*(cpSource + iCounter) != 0)
        { ok = 0; break; }

   if (ok)
   {
     // Copies the variable (binary)
     memcpy(cpDest, cpSource, BL_min(iSourceSize, iDestSize));

     // Byte reversal (little endian -> big endian)
     ByteRev(cpDest, iDestSize);
   }

  #else // BIG endian

   // Checks if variable fits in the model
   if (iSourceSize>iDestSize)
     for (iCounter = 0; iCounter<iSourceSize-iDestSize; iCounter++)
      if (*(cpSource + iCounter) != 0)
        { ok = 0; break; }

   if (ok)
   {
     // Copies the variable (binary)

     if (iSourceSize>iDestSize)
     {
       iCounter = iSourceSize - iDestSize;
       memcpy(cpDest, cpSource + iCounter, iDestSize);
     }
     else
     {
       iCounter = iDestSize - iSourceSize;
       memcpy(cpDest + iCounter, cpSource, iSourceSize);
     }

   }

  #endif // BIG endian

  return (ok) ? BL_CONV_OK : BL_CANT_SHRINK;
}


int
PutInt (char * cpSource, char * cpDest, int iSourceSize, int iDestSize)
{
  //
  // Function to degrade a signed integer of a certain size to a known
  // size in big endian format used by BLlib.
  // The value is read from cpSource and stored in cpDest.
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: ByteRev;
  //
  // Parameters:   char * cpDest        pointer to dest. (already allocated)
  //               char * cpSource      pointer to source
  //               int    iSourceSize   source size (bytes)
  //               int    iDestSize     destination size (bytes)
  //
  // Returns:      BL_CONV_OK     if transformation succesfull
  //               BL_CANT_SHRINK if transformation failed
  //
  // Note:
  //               cpDest is set to all zeroes if degrading
  //               was not possible.
  //


  register int iCounter;
  register unsigned char ucFillByte;
  register char ok = 1;

  #ifdef BL_LITTLE_END

  ucFillByte = (*(cpSource + iSourceSize - 1) & 0x80) ? 0xff : 0;

  #else  // BIG endian

  ucFillByte = (*(cpSource) & 0x80) ? 0xff : 0;

  #endif // BIG endian

  memset(cpDest, ucFillByte, iDestSize);

  #ifdef BL_LITTLE_END

   // Checks if variable fits in the model
   if (iSourceSize>iDestSize)
   {
     for (iCounter = iDestSize; iCounter<iSourceSize; iCounter++)
      if (((unsigned char) *(cpSource + iCounter)) != ucFillByte)
        { ok = 0; break; }

     // If ok, now degrading is possible on the following conditions:
     // 1) source is negative and most significant bit of most
     //    significant byte remaining is 1;
     // 2) source is positive and most significant bit of most
     //    significant byte remaining is 0.

     if ( ((*(cpSource + iDestSize - 1) & 0x80) && !ucFillByte) ||
          (!(*(cpSource + iDestSize - 1) & 0x80) && ucFillByte) )
       ok = 0;
   }

   if (ok)
   {
     // Copies the variable (binary)
     memcpy(cpDest, cpSource, BL_min(iSourceSize,iDestSize));

     // Byte reversal (little endian -> big endian)
     ByteRev(cpDest, iDestSize);

   }
   else
     memset(cpDest, 0, iDestSize); // Conversion failed

  #else // BIG endian

   // Checks if variable fits in the model
   if (iSourceSize>iDestSize)
   {
     for (iCounter = 0; iCounter<iSourceSize-iDestSize; iCounter++)
      if (((unsigned char) *(cpSource + iCounter)) != ucFillByte)
        { ok = 0; break; }

     // If ok, now degrading is possible on the following conditions:
     // 1) source is negative and most significant bit of most
     //    significant byte remaining is 1;
     // 2) source is positive and most significant bit of most
     //    significant byte remaining is 0.

     if ( ((*(cpSource + iSourceSize - iDestSize) & 0x80) && !ucFillByte) ||
          (!(*(cpSource + iSourceSize - iDestSize) & 0x80) && ucFillByte) )
       ok = 0;
   }

   if (ok)
   {
     // Copies the variable (binary)

     if (iSourceSize>iDestSize)
     {
       iCounter = iSourceSize - iDestSize;
       memcpy(cpDest, cpSource + iCounter, iDestSize);
     }
     else
     {
       iCounter = iDestSize - iSourceSize;
       memcpy(cpDest + iCounter, cpSource, iSourceSize);
     }

   }
   else
     memset(cpDest, 0, iDestSize); // Conversion failed

  #endif // BIG endian

  return (ok) ? BL_CONV_OK : BL_CANT_SHRINK;
}


// Functions used to revert an integer from the BLlib memory format
// to the machine format.


int
GetUintInUint (char * cpSource, char * cpDest, int iSourceSize, int iDestSize)
{
  //
  // Function to transform an unsigned integer in the BLlib format
  // into the corresponding value in the internal representation.
  // The value is read from cpSource and stored in cpDest.
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: ByteRev;
  //
  // Parameters:   char * cpDest        pointer to dest. (already allocated)
  //               char * cpSource      pointer to source
  //               int    iSourceSize   source size (bytes)
  //               int    iDestSize     destination size (bytes)
  //
  // Returns:      BL_CONV_OK     if transformation succesfull
  //               BL_CANT_SHRINK if transformation failed
  //
  // Note:
  //               cpDest is set to all zeroes if transformation
  //               failed.
  //


  register int iCounter;
  register char ok = 1;

  memset(cpDest, 0, iDestSize);

  // Checks if variable fits in the model
  if (iSourceSize>iDestSize)
    for (iCounter = 0; iCounter<iSourceSize-iDestSize; iCounter++)
     if (*(cpSource + iCounter) != 0)
       { ok = 0; break; }

  if (ok)
  {
    // Copies the variable (binary)

    if (iSourceSize>iDestSize)
    {
      iCounter = iSourceSize - iDestSize;
      memcpy(cpDest, cpSource + iCounter, iDestSize);
    }
    else
    {
      iCounter = iDestSize - iSourceSize;
      memcpy(cpDest + iCounter, cpSource, iSourceSize);
    }

    #ifdef BL_LITTLE_END

      // Byte reversal (big endian -> little endian)
      ByteRev(cpDest, iDestSize);

    #endif // BL_LITTLE_END

  }

  return (ok) ? BL_CONV_OK : BL_CANT_SHRINK;

}


int
GetIntInInt (char * cpSource, char * cpDest, int iSourceSize, int iDestSize)
{
  //
  // Function to transform a signed integer in the BLlib format
  // into the corresponding value in the internal representation.
  // The value is read from cpSource and stored in cpDest.
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: ByteRev;
  //
  // Parameters:   char * cpDest        pointer to dest. (already allocated)
  //               char * cpSource      pointer to source
  //               int    iSourceSize   source size (bytes)
  //               int    iDestSize     destination size (bytes)
  //
  // Returns:      BL_CONV_OK     if transformation succesfull
  //               BL_CANT_SHRINK if transformation failed
  //
  // Note:
  //               cpDest is set to all zeroes if transformation
  //               failed.
  //


  register int iCounter;
  register char ok = 1;
  unsigned char ucFillByte;

  ucFillByte = (*(cpSource) & 0x80) ? 0xff : 0;

  memset(cpDest, ucFillByte, iDestSize);

  // Checks if variable fits in the model
  if (iSourceSize>iDestSize)
  {
    for (iCounter = 0; iCounter<iSourceSize-iDestSize; iCounter++)
     if (((unsigned char) *(cpSource + iCounter)) != ucFillByte)
       { ok = 0; break; }

    // If ok, now conversion is possible on the following conditions:
    // 1) source is negative and most significant bit of most
    //    significant byte remaining is 1;
    // 2) source is positive and most significant bit of most
    //    significant byte remaining is 0.

    if ( ((*(cpSource + iSourceSize - iDestSize) & 0x80) && !ucFillByte) ||
         (!(*(cpSource + iSourceSize - iDestSize) & 0x80) && ucFillByte) )
      ok = 0;
  }

  if (ok)
  {
    // Copies the variable (binary)

    if (iSourceSize>iDestSize)
    {
      iCounter = iSourceSize - iDestSize;
      memcpy(cpDest, cpSource + iCounter, iDestSize);
    }
    else
    {
      iCounter = iDestSize - iSourceSize;
      memcpy(cpDest + iCounter, cpSource, iSourceSize);
    }

    #ifdef BL_LITTLE_END

      // Byte reversal (big endian -> little endian)
      ByteRev(cpDest, iDestSize);

    #endif // BL_LITTLE_END

  }
  else
    memset(cpDest, 0, iDestSize); // Conversion failed

  return (ok) ? BL_CONV_OK : BL_CANT_SHRINK;

}


int
GetIntInUint (char * cpSource, char * cpDest, int iSourceSize, int iDestSize)
{
  //
  // Function to transform a signed integer in the BLlib format
  // into the corresponding value (unsigned) in the internal
  // representation.
  // The value is read from cpSource and stored in cpDest.
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: ByteRev;
  //
  // Parameters:   char * cpDest        pointer to dest. (already allocated)
  //               char * cpSource      pointer to source
  //               int    iSourceSize   source size (bytes)
  //               int    iDestSize     destination size (bytes)
  //
  // Returns:      BL_CONV_OK                    if transformation succesfull
  //               BL_CANT_SHRINK/BL_NEG_TO_POS  if transformation failed
  //
  // Note:
  //               cpDest is set to all zeroes if transformation
  //               failed.
  //


  register int iCounter, iResult = BL_CANT_SHRINK;
  register char ok = 1;
  unsigned char ucFillByte;

  ucFillByte = (*(cpSource) & 0x80) ? 0xff : 0;

  memset(cpDest, ucFillByte, iDestSize);

  // Checks if variable fits in the model

  if (iSourceSize>iDestSize)
  {
    for (iCounter = 0; iCounter<iSourceSize-iDestSize; iCounter++)
     if (((unsigned char) *(cpSource + iCounter)) != ucFillByte)
       { ok = 0; break; }

    // If ok, now conversion is possible if source is positive.


    if (ok)
      if (ucFillByte)
      {
        iResult = BL_NEG_TO_POS;
        ok = 0;
      }
  }

  if (ok)
  {
    // Copies the variable (binary)

    if (iSourceSize>iDestSize)
    {
      iCounter = iSourceSize - iDestSize;
      memcpy(cpDest, cpSource + iCounter, iDestSize);
    }
    else
    {
      iCounter = iDestSize - iSourceSize;
      memcpy(cpDest + iCounter, cpSource, iSourceSize);
    }

    #ifdef BL_LITTLE_END

      // Byte reversal (big endian -> little endian)
      ByteRev(cpDest, iDestSize);

    #endif // BL_LITTLE_END

  }
  else
    memset(cpDest, 0, iDestSize); // Conversion failed

  return (ok) ? BL_CONV_OK : iResult;

}


int
GetUintInInt (char * cpSource, char * cpDest, int iSourceSize, int iDestSize)
{
  //
  // Function to transform an unsigned integer in the BLlib format into
  // the corresponding value (signed) in the internal representation.
  // The value is read from cpSource and stored in cpDest.
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: ByteRev;
  //
  // Parameters:   char * cpDest        pointer to dest. (already allocated)
  //               char * cpSource      pointer to source
  //               int    iSourceSize   source size (bytes)
  //               int    iDestSize     destination size (bytes)
  //
  // Returns:      BL_CONV_OK      if transformation succesfull
  //               BL_CANT_SHRINK  if transformation failed
  //
  // Note:
  //               cpDest is set to all zeroes if transformation
  //               failed.
  //


  register int iCounter;
  register char ok = 1;

  memset(cpDest, 0, iDestSize);

  // Checks if variable fits in the model
  if (iSourceSize>iDestSize)
  {
    for (iCounter = 0; iCounter<iSourceSize-iDestSize; iCounter++)
     if (((unsigned char) *(cpSource + iCounter)) != 0)
       { ok = 0; break; }

    // If ok, now conversion is possible if most significant bit of most
    // significant byte remaining is 0.


    if ( (*(cpSource + iSourceSize - iDestSize) & 0x80) )
      ok = 0;
  }

  if (ok)
  {
    // Copies the variable (binary)

    if (iSourceSize>iDestSize)
    {
      iCounter = iSourceSize - iDestSize;
      memcpy(cpDest, cpSource + iCounter, iDestSize);
    }
    else
    {
      iCounter = iDestSize - iSourceSize;
      memcpy(cpDest + iCounter, cpSource, iSourceSize);
    }

    #ifdef BL_LITTLE_END

      // Byte reversal (big endian -> little endian)
      ByteRev(cpDest, iDestSize);

    #endif // BL_LITTLE_END

  }
  else
    memset(cpDest, 0, iDestSize); // Conversion failed

  return (ok) ? BL_CONV_OK : BL_CANT_SHRINK;

}

// Templates for performing automatic conversions from integers

template <class T> int
FromIntInInt(char * cpSource, int iSourceSize, T &var)
{
  return GetIntInInt(cpSource, (char *) (&var), iSourceSize, sizeof(var));
}


template <class T> int
FromUintInUint(char * cpSource, int iSourceSize, T &var)
{
  return GetUintInUint(cpSource, (char *) (&var), iSourceSize, sizeof(var));
}


template <class T> int
FromIntInUint(char * cpSource, int iSourceSize, T &var)
{
  return GetIntInUint(cpSource, (char *) (&var), iSourceSize, sizeof(var));
}


template <class T> int
FromUintInInt(char * cpSource, int iSourceSize, T &var)
{
  return GetUintInInt(cpSource, (char *) (&var), iSourceSize, sizeof(var));
}


template <class T> int
ToInt(char * cpDest, int iDestSize, T &var)
{
  return PutInt((char *) (&var), cpDest, sizeof(var), iDestSize);
}


template <class T> int
ToUint(char * cpDest, int iDestSize, T &var)
{
  return PutUint((char *) (&var), cpDest, sizeof(var), iDestSize);
}


//
// BL_ToIntxx / BL_ToUintxx functions (using BL_ixx / BL_uixx public variables)
//
// If the user is not sure that the value passed to the BL_ToIntxx / BL_ToUintxx
// function fits in the respective memory model, the function should be
// called alone and then the value of BL_convResult should be tested
// for equality with the constant value BL_CONV_OK; if the conversion was
// succesfull, now the user can send the value accessing the correct
// public variable.
//
// e.g.
//
// BL_Message MyMessage;             // Creates a message to send
//
// BL_ToInt16(12345);                // Converts the value to BL_Int16 type
// if (BL_convResult == BL_CONV_OK)  // Conversion ok
//   MyMessage << BL_i16;            // Writes the value obtained to MyMessage
//


BL_Uint16
BL_ToUint16(unsigned long ulNumber)
{
  BL_convResult = ToUint((char *) &(BL_ui16.value), 2, ulNumber);
  return BL_ui16;
}


BL_Int16
BL_ToInt16(long lNumber)
{
  BL_convResult = ToInt((char *) &(BL_i16.value), 2, lNumber);
  return BL_i16;
}


BL_Uint32
BL_ToUint32(unsigned long ulNumber)
{
  BL_convResult = ToUint((char *) &(BL_ui32.value), 4, ulNumber);
  return BL_ui32;
}


BL_Int32
BL_ToInt32(long lNumber)
{
  BL_convResult = ToInt((char *) &(BL_i32.value), 4, lNumber);
  return BL_i32;
}


BL_Uint64
BL_ToUint64(unsigned long ulNumber)
{
  BL_convResult = ToUint((char *) &(BL_ui64.value), 8, ulNumber);
  return BL_ui64;
}


BL_Int64
BL_ToInt64(long lNumber)
{
  BL_convResult = ToInt((char *) &(BL_i64.value), 8, lNumber);
  return BL_i64;
}


//******************************************************************
//************* FLOATING POINT MANIPULATION FUNCTIONS **************
//******************************************************************


#ifdef BL_FP_IEEE_754

 // Functions used to convert/revert a real to/from the BLlib memory
 // format.


 void
 PutReal (char * cpSource, char * cpDest, int iSourceSize)
 {

  //
  // Function to put/get a real in the message; the real is always stored
  // in the big endian format used by BLlib.
  // The value is read from cpSource and stored in cpDest.
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: ByteRev;
  //
  // Parameters:   char * cpDest        pointer to dest. (already allocated)
  //               char * cpSource      pointer to source
  //               int    iSourceSize   source size (bytes)
  //
  // Returns:      void
  //


  memcpy(cpDest, cpSource, iSourceSize);

  #ifdef BL_LITTLE_END

   ByteRev(cpDest, iSourceSize);

  #endif //BL_LITTLE_END

  return;

 }


 int
 DoubleToFloat (char * cpDouble, float &fValue)
 {

  //
  // Function to transform a floting point double precision into
  // a floating point single precision, both in standard IEEE 754.
  // The value is read from cpDouble (big endian format) and stored
  // in cpFloat (machine dependent format).
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: ByteRev;
  //
  // Parameters:   char * cpDouble      pointer to double
  //               char * cpFloat       pointer to float
  //
  // Returns:
  //               BL_PREC_LOSS    if transformation succesfull
  //               BL_CANT_SHRINK  if transformation failed
  //


  double dNewDouble;

  memcpy((char *) &dNewDouble, cpDouble, _BL_DOUBLE_SIZE);

  #ifdef BL_LITTLE_END

    // Byte reversal (big endian -> little endian)
    ByteRev((char *) &dNewDouble, _BL_DOUBLE_SIZE);

  #endif // BL_LITTLE_END

  if ((dNewDouble > FLT_MAX) || (dNewDouble < FLT_MIN))
    return BL_CANT_SHRINK;

  fValue = dNewDouble;

  return BL_PREC_LOSS;

 }


 template <class T> int
 RealToInt(char *cpSource, char cSourceType, T &var, TypeOfIntegers iDestType)
 {
  //
  // Function template to convert a real (float or double) to an integer whose
  // type is passed in iDestType.
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: PutReal;
  //
  // Parameters:   char *cpSource        pointer to item data in buffer
  //               cSourceType           item's type (float or double)
  //               T     &Var            integer number (by reference)
  //               int   iDestType       type of integer
  //
  // Returns:      BL_FP_TO_INT          conversion succesfull
  //               BL_NEG_TO_POS         tried to extract a negative value
  //                                     to an unsigned one
  //               BL_CANT_SHRINK        type compatibles, size incompatibles
  //
  // Note:         var overwritten only if conversion succesfull
  //


  double dValue, dMin, dMax;
  float  fValue;

  if (cSourceType == BL_FLOAT) // Float stored in a double
  {

    PutReal(cpSource, (char *) &fValue, _BL_FLOAT_SIZE);
    dValue = (double) fValue;

  }

  else // Double stored in a double

    PutReal(cpSource, (char *) &dValue, _BL_DOUBLE_SIZE);

  dValue = floor(dValue);

  switch (iDestType) // Determines min & max values for integer
  {
    case _BL_CHAR:
      dMin = (double) SCHAR_MIN;
      dMax = (double) SCHAR_MAX;
      break;
    case _BL_SHORT:
      dMin = (double) SHRT_MIN;
      dMax = (double) SHRT_MAX;
      break;
    case _BL_INT:
      dMin = (double) INT_MIN;
      dMax = (double) INT_MAX;
      break;
    case _BL_LONG:
      dMin = (double) LONG_MIN;
      dMax = (double) LONG_MAX;
      break;
    default:
      dMin = (double) 0;
      switch (iDestType)
      {
        case _BL_UCHAR:
          dMax = (double) UCHAR_MAX;
          break;
        case _BL_USHORT:
          dMax = (double) USHRT_MAX;
          break;
        case _BL_UINT:
          dMax = (double) UINT_MAX;
          break;
        case _BL_ULONG:
          dMax = (double) ULONG_MAX;
          break;
      }
      break;
  }

  if ((dValue >= dMin) && (dValue <= dMax))
  {
    var = (T) dValue;
    return BL_FP_TO_INT;
  }
  else
    if ((iDestType > _BL_LONG) && (dValue < 0.0))
      return BL_NEG_TO_POS;
    else
      return BL_CANT_SHRINK;
 }


 template <class T> int
 IntToReal(BL_Message &msg, char cType, T &var)
 {
  //
  // Function template to extract an integer, whose exact type is passed
  // in cType, to a fp (single or double precision).
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: PutReal;
  //
  // Parameters:   BL_Message &msg        message to operate with
  //               char    cType         type of integer
  //               T       &Var          float number (by reference)
  //
  // Returns:      BL_INT_TO_FP          (conversion can't fail)
  //


  unsigned long ulCurrentItem, ulValue;
  long lValue;

  ulCurrentItem = msg.SkipItem(0); // Saves current position

  if (cType >= BL_UCHAR)
  {
    msg >> ulValue;
    var = (T) ulValue;
  }
  else
  {
    msg >> lValue;
    var = (T) lValue;
  }

  msg.SeekItem(ulCurrentItem);     // Return to old position

  return BL_INT_TO_FP;

 }

#endif // BL_FP_IEEE_754


//******************************************************************
//********************* CLASS MESSAGE METHODS **********************
//******************************************************************


BL_Message::BL_Message(BL_Message &ForwMail)
{
  // Copy constructor

  ulBufOffset = 0;
  ulCurrentItem = 1;
  i32Type = ForwMail.i32Type;
  ulDataLen = ForwMail.ulDataLen;
  if (cpMsgBuf != NULL)
    free(cpMsgBuf);
  BL_ChkNul(cpMsgBuf = (char *) malloc(ulDataLen));
  memcpy(cpMsgBuf, ForwMail.cpMsgBuf, ulDataLen);
}


void
BL_Message::Clear ()
{
   // Method to clean up a message

   i32Type = BL_NOT_DEFINED;
   ulDataLen = 0;
   if (cpMsgBuf != NULL)
     free(cpMsgBuf);
   cpMsgBuf = NULL;
   ulBufOffset = 0;
   ulCurrentItem = 1;
}


long
BL_Message::SkipItem(long lItemDisplacement)
{
  //
  // BL_Message method to skip onward/backward a certain number of items.
  // A positive value for iItemDisplacement move the item
  // pointer towards the end of the buffer; a negative value
  // moves the pointer towards the beginning of the buffer.
  // A null value for iItemDisplacement returns the current
  // item number.
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: CalcItemSize;
  //
  // Parameters:   long lItemDisplacement    number of items to skip
  //
  // Returns:      number of really skipped items (lItemDisplacement!=0)
  //               (always positive)
  //           or
  //               current item number (lItemDisplacement==0)
  //


  register unsigned long ulAbsItemNum, ulSkippedItems = 0, ulOldItem;


  if (lItemDisplacement == 0)
    return ulCurrentItem;           // Function used in info mode

  if (lItemDisplacement < 0)

    if ((lItemDisplacement * (-1)) > ulCurrentItem)

      ulAbsItemNum = 0;

    else

      ulAbsItemNum = ulCurrentItem + lItemDisplacement;

  else

    ulAbsItemNum = ulCurrentItem + lItemDisplacement;

  if (ulAbsItemNum == 0)            // Illegal back move (pointer before start)
  {
    ulSkippedItems = ulCurrentItem-1;
    ulCurrentItem = 1;
    ulBufOffset = 0;
    return ulSkippedItems;
  }

  ulOldItem = ulCurrentItem;

  if (ulAbsItemNum < ulCurrentItem) // Legal back move
  {
    ulBufOffset = 0;
    ulCurrentItem = 1;
  }

  while (ulCurrentItem != ulAbsItemNum)
  {

    ulBufOffset += CalcItemSize(cpMsgBuf + ulBufOffset);
    ulCurrentItem++;
    if (ulBufOffset > ulDataLen)   // Pointer gone too far
    {
      ulBufOffset = ulDataLen;
      break;
    }

  }

  return ( (ulCurrentItem > ulOldItem) ? (ulCurrentItem - ulOldItem)
          : (ulOldItem - ulCurrentItem) );

}


unsigned long
BL_Message::SeekItem(unsigned long ulAbsItemNum)
{
  //
  // BL_Message method to position the item pointer to the item whose
  // number is passed in ulAbsItemNum.
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: CalcItemSize;
  //
  // Parameters:   ulong ulAbsItemNum        item number
  //
  // Returns:      new item number
  //


  if (ulAbsItemNum <= 1)            // First item
  {
    ulCurrentItem = 1;
    ulBufOffset = 0;
    return ulCurrentItem;
  }

  if (ulAbsItemNum < ulCurrentItem) // Legal back move
  {
    ulBufOffset = 0;
    ulCurrentItem = 1;
  }

  while (ulCurrentItem != ulAbsItemNum)
  {

    ulBufOffset += CalcItemSize(cpMsgBuf + ulBufOffset);
    ulCurrentItem++;
    if (ulBufOffset > ulDataLen)    // Pointer gone too far
    {
      ulBufOffset = ulDataLen;
      break;
    }

  }

  return ulCurrentItem;

}


int
BL_Message::DeleteItem(unsigned long ulAbsItemNum)
{
  //
  // BL_Message method to delete the item corresponding to the number
  // passed in ulAbsItemNum. After this operation, the item is lost
  // and can not be accessed anymore.
  // Item pointer, after operation, remains over the same item where
  // it was before; if this item was the one deleted, the pointer
  // is set to the next item (if exists).
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: CalcItemSize;
  //
  // Parameters:   ulong ulAbsItemNum     item number
  //
  // Returns:      BL_ITEM_DELETED        if item has been deleted
  //               BL_WRONG_ITEM          otherwise
  //


  register unsigned long ulOldItem, ulOldBufOffset, ulItemSize;

  if (ulAbsItemNum == 0)            // Illegal item specified
    return BL_WRONG_ITEM;

  ulOldItem = ulCurrentItem;
  ulOldBufOffset = ulBufOffset;

  if (ulAbsItemNum < ulCurrentItem) // Legal back move
  {
    ulBufOffset = 0;
    ulCurrentItem = 1;
  }

  while (ulCurrentItem != ulAbsItemNum)
  {
    ulBufOffset += CalcItemSize(cpMsgBuf + ulBufOffset);
    ulCurrentItem++;
    if (ulBufOffset > ulDataLen)    // Pointer gone too far
    {
      ulBufOffset = ulDataLen;
      break;
    }
  }

  if ( (ulCurrentItem == ulAbsItemNum) && (ulBufOffset != ulDataLen) )
  {
    // Item specified exists
    ulItemSize = CalcItemSize(cpMsgBuf + ulBufOffset);

    if ( (ulDataLen - ulItemSize - ulBufOffset) > 0 ) // Not last item
      memcpy(cpMsgBuf + ulBufOffset, cpMsgBuf + ulBufOffset + ulItemSize,
             ulDataLen - ulItemSize - ulBufOffset);

    ulDataLen -= ulItemSize;

    cpMsgBuf = (char *) realloc(cpMsgBuf, ulDataLen);


    if (ulOldItem > ulCurrentItem)
    {
      ulCurrentItem = ulOldItem - 1;
      ulBufOffset = ulOldBufOffset - ulItemSize;
    }
    else
      if (ulOldItem == ulCurrentItem)
        ulBufOffset = BL_min(ulOldBufOffset, ulDataLen);
      else
      {
        ulCurrentItem = ulOldItem;
        ulBufOffset = ulOldBufOffset;
      }

    return BL_ITEM_DELETED;
  }
  else
  {
    ulCurrentItem = ulOldItem;
    ulBufOffset = ulOldBufOffset;
    return BL_WRONG_ITEM;
  }
}


//******************************************************************
//********************** INSERTION OPERATORS ***********************
//******************************************************************

// Here follows the family of overloaded insertion operators on class
// BL_Message.
// Each operator works on the object this way:
// 1) reallocates memory for cpMsgBuf, so that it can store the old
//    cpMsgBuf and the new data;
// 2) appends the base type signature;
// 3) appends the size of following data, if necessary;
// 4) appends the data, purposely formatted.
// If the message is made up of multiple data items, the user, after
// putting in the last item with one of these operators, must call the method
// SetType to set the type tag to a value known by both the sender and
// receiver of the message; it is the task of programmer to use the
// notation carefully and accurately.
// The data items in the message are stored sequentially using the model:
// type-[dimension]-data. The dimension field applies only to variable-
// length types, such as strings & rawdata.
// The data signatures are as follows:
// _CHAR     signed char
// _UCHAR    unsigned char
// _INT16    16 bit signed integers
// _UINT16   16 bit unsigned integers
// _INT32    32 bit signed integers
// _UINT32   32 bit unsigned integers
// _INT64    64 bit signed integers
// _UINT64   64 bit unsigned integers
// _FLOAT    floating point single precision
// _DOUBLE   floating point double precision
// _CHARPT   tiny string null terminated (2^8-1 bytes max - 1 dimension byte)
// _CHARPS   small string null terminated (2^16-1 bytes max - 2 dimension bytes)
// _CHARPL   large string null terminated (2^24-1 bytes max - 3 dimension bytes)
// _CHARPH   huge string null terminated (2^32-1 bytes max - 4 dimension bytes)
// _RDATAT   tiny binary data (2^8-1 bytes max - 1 bytes for dimension)
// _RDATAS   small binary data (2^16-1 bytes max - 2 bytes for dimension)
// _RDATAL   large binary data (2^24-1 bytes max - 3 bytes for dimension)
// _RDATAH   huge binary data (2^32-1 bytes max - 4 bytes for dimension)
//
// Thus, to send an integer with a precise memory usage, machine
// independent, the user must specify the appropriate conversion
// function: BL_ToIntxx(), BL_ToUintxx().
//
// E.g. msg << BL_ToInt16(123)
//
// Each call to BL_ToIntxx overwrites the contents of the public variable
// BL_ixx, which is then accessed to retrieve the data.
// Integers are sent using the same number of bytes of the type of the
// variable if the user does not request a precise memory model.
// All numeric values are stored in the message in the Big Endian format.


#ifdef BL_FP_IEEE_754


 BL_Message&
 BL_Message::operator<<(float fNumber)
 {

  //
  // BL_Message method to put a standard IEEE 754 single precision floating
  // point in the message; the fp is always stored in the big endian
  // format used by BLlib.
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: ResizeMsg;
  //               PutReal;
  //
  // Parameters:   float  fNumber       floating point number
  //
  // Returns:      the modified BL_Message
  //


  ResizeMsg(cpMsgBuf, ulDataLen + _BL_SIGN_SIZE + _BL_FLOAT_SIZE,
            ulDataLen, ulBufOffset);
  *(cpMsgBuf + ulDataLen) = _FLOAT;
  PutReal((char *) &fNumber, cpMsgBuf + ulDataLen + _BL_SIGN_SIZE,
          _BL_FLOAT_SIZE);
  ulDataLen += (_BL_SIGN_SIZE + _BL_FLOAT_SIZE);
  BL_convResult = BL_CONV_OK;
  return *this;

 }


 BL_Message&
 BL_Message::operator<<(double dNumber)
 {

  //
  // BL_Message method to put a standard IEEE 754 double precision floating
  // point in the message; the fp is always stored in the big endian
  // format used by BLlib.
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: ResizeMsg;
  //               PutReal;
  //
  // Parameters:   double  dNumber       floating point number
  //
  // Returns:      the modified BL_Message
  //


  ResizeMsg(cpMsgBuf, ulDataLen + _BL_SIGN_SIZE + _BL_DOUBLE_SIZE,
            ulDataLen, ulBufOffset);
  *(cpMsgBuf + ulDataLen) = _DOUBLE;
  PutReal((char *) &dNumber, cpMsgBuf + ulDataLen + _BL_SIGN_SIZE,
          _BL_DOUBLE_SIZE);
  ulDataLen += (_BL_SIGN_SIZE + _BL_DOUBLE_SIZE);
  BL_convResult = BL_CONV_OK;
  return *this;

 }

#endif // BL_FP_IEEE_754


BL_Message&
BL_Message::operator<<(char cCharacter)
{

  //
  // BL_Message method to put a signed character in the message.
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: ResizeMsg;
  //
  // Parameters:   char  cCharacter       character value
  //
  // Returns:      the modified BL_Message
  //


  ResizeMsg(cpMsgBuf, ulDataLen + _BL_SIGN_SIZE + _BL_CHAR_SIZE,
            ulDataLen, ulBufOffset);
  *(cpMsgBuf + ulDataLen) = _CHAR;
  *(cpMsgBuf + ulDataLen + _BL_SIGN_SIZE) = cCharacter;
  ulDataLen += (_BL_SIGN_SIZE + _BL_CHAR_SIZE);
  return *this;

}


BL_Message&
BL_Message::operator<<(unsigned char ucCharacter)
{

  //
  // BL_Message method to put an unsigned character in the message.
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: ResizeMsg;
  //
  // Parameters:   uchar  ucCharacter       character value
  //
  // Returns:      the modified BL_Message
  //


  ResizeMsg(cpMsgBuf, ulDataLen + _BL_SIGN_SIZE + _BL_CHAR_SIZE,
            ulDataLen, ulBufOffset);
  *(cpMsgBuf + ulDataLen) = _UCHAR;
  *(cpMsgBuf + ulDataLen + _BL_SIGN_SIZE) = ucCharacter;
  ulDataLen += (_BL_SIGN_SIZE + _BL_CHAR_SIZE);
  return *this;

}


BL_Message&
BL_Message::operator<<(short sNumber)
{

  //
  // BL_Message method to put a signed short integer in the message.
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: BL_Message<<(BL_Int16);
  //               BL_Message<<(BL_Int32);
  //               BL_ToInt16;
  //               BL_ToInt32;
  //
  // Parameters:   short  sNumber           short value
  //
  // Returns:      the modified BL_Message
  //
  // Note:         assumes that short is at least 2 bytes long and
  //               at most 4 bytes long
  //


  // Decides which integer model to use
  switch (sizeof(sNumber))
  {
    case 2:
      return *this<<BL_ToInt16(sNumber);
    case 4:
      return *this<<BL_ToInt32(sNumber);
    default:
      return *this;
  }

}


BL_Message&
BL_Message::operator<<(unsigned short usNumber)
{

  //
  // BL_Message method to put an unsigned short integer in the message.
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: BL_Message<<(BL_Uint16);
  //               BL_Message<<(BL_Uint32);
  //               BL_ToUint16;
  //               BL_ToUint32;
  //
  // Parameters:   ushort  usNumber         unsigned short value
  //
  // Returns:      the modified BL_Message
  //
  // Note:         assumes that unsigned short is at least 2 bytes long
  //               and at most 4 bytes long
  //


  // Decides which integer model to use
  switch (sizeof(usNumber))
  {
    case 2:
      return *this<<BL_ToUint16(usNumber);
    case 4:
      return *this<<BL_ToUint32(usNumber);
    default:
      return *this;
  }

}


BL_Message&
BL_Message::operator<<(int iNumber)
{

  //
  // BL_Message method to put a signed integer in the message.
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: BL_Message<<(BL_Int16);
  //               BL_Message<<(BL_Int32);
  //               BL_Message<<(BL_Int64);
  //               BL_ToInt16;
  //               BL_ToInt32;
  //               BL_ToInt64;
  //
  // Parameters:   int iNumber              int value
  //
  // Returns:      the modified BL_Message
  //
  // Note:         assumes that int is at least 2 bytes long
  //               and at most 8 bytes long
  //


  // Decides which integer model to use
  switch (sizeof(iNumber))
  {
    case 2:
      return *this<<BL_ToInt16(iNumber);
    case 4:
      return *this<<BL_ToInt32(iNumber);
    case 8:
      return *this<<BL_ToInt64(iNumber);
    default:
      return *this;
  }

}


BL_Message&
BL_Message::operator<<(unsigned int uiNumber)
{

  //
  // BL_Message method to put an unsigned integer in the message.
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: BL_Message<<(BL_Uint16);
  //               BL_Message<<(BL_Uint32);
  //               BL_Message<<(BL_Uint64);
  //               BL_ToUint16;
  //               BL_ToUint32;
  //               BL_ToUint64;
  //
  // Parameters:   uint uiNumber            unsigned int value
  //
  // Returns:      the modified BL_Message
  //
  // Note:         assumes that unsigned int is at least 2 bytes long
  //               and at most 8 bytes long
  //


  // Decides which integer model to use
  switch (sizeof(uiNumber))
  {
    case 2:
      return *this<<BL_ToUint16(uiNumber);
    case 4:
      return *this<<BL_ToUint32(uiNumber);
    case 8:
      return *this<<BL_ToUint64(uiNumber);
    default:
      return *this;
  }

}


BL_Message&
BL_Message::operator<<(long lNumber)
{

  //
  // BL_Message method to put a signed long integer in the message.
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: BL_Message<<(BL_Int32);
  //               BL_Message<<(BL_Int64);
  //               BL_ToInt32;
  //               BL_ToInt64;
  //
  // Parameters:   long lNumber             long value
  //
  // Returns:      the modified BL_Message
  //
  // Note:         assumes that long is at least 4 bytes long
  //               and at most 8 bytes long
  //


  // Decides which integer model to use
  switch (sizeof(lNumber))
  {
    case 4:
      return *this<<BL_ToInt32(lNumber);
    case 8:
      return *this<<BL_ToInt64(lNumber);
    default:
      return *this;
  }

}


BL_Message&
BL_Message::operator<<(unsigned long ulNumber)
{

  //
  // BL_Message method to put an unsigned long integer in the message.
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: BL_Message<<(BL_Uint32);
  //               BL_Message<<(BL_Uint64);
  //               BL_ToUint32;
  //               BL_ToUint64;
  //
  // Parameters:   ulong ulNumber           unsigned long value
  //
  // Returns:      the modified BL_Message
  //
  // Note:         assumes that unsigned long is at least 4 bytes long
  //               and at most 8 bytes long
  //


  // Decides which integer model to use
  switch (sizeof(ulNumber))
  {
    case 4:
      return *this<<BL_ToUint32(ulNumber);
    case 8:
      return *this<<BL_ToUint64(ulNumber);
    default:
      return *this;
  }

}


BL_Message&
BL_Message::operator<<(BL_Int16 i16Number)
{

  //
  // BL_Message method to put an BL_Int16 integer in the message.
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: ResizeMsg;
  //
  // Parameters:   BL_Int16 i16Number          BL_Int16 value
  //
  // Returns:      the modified BL_Message
  //


  ResizeMsg(cpMsgBuf, ulDataLen + _BL_SIGN_SIZE + 2,
            ulDataLen, ulBufOffset);
  *(cpMsgBuf + ulDataLen) = _INT16;
  memcpy(cpMsgBuf + ulDataLen + _BL_SIGN_SIZE, (char *) &(i16Number.value), 2);
  ulDataLen += (_BL_SIGN_SIZE + 2);
  return *this;

}


BL_Message&
BL_Message::operator<<(BL_Uint16 ui16Number)
{

  //
  // BL_Message method to put an BL_Uint16 integer in the message.
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: ResizeMsg;
  //
  // Parameters:   BL_Uint16 ui16Number        BL_Uint16 value
  //
  // Returns:      the modified BL_Message
  //


  ResizeMsg(cpMsgBuf, ulDataLen + _BL_SIGN_SIZE + 2,
            ulDataLen, ulBufOffset);
  *(cpMsgBuf + ulDataLen) = _UINT16;
  memcpy(cpMsgBuf + ulDataLen + _BL_SIGN_SIZE, (char *) &(ui16Number.value), 2);
  ulDataLen += (_BL_SIGN_SIZE + 2);
  return *this;

}


BL_Message&
BL_Message::operator<<(BL_Int32 i32Number)
{

  //
  // BL_Message method to put an BL_Int32 integer in the message.
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: ResizeMsg;
  //
  // Parameters:   BL_Int32 i32Number          BL_Int32 value
  //
  // Returns:      the modified BL_Message
  //


  ResizeMsg(cpMsgBuf, ulDataLen + _BL_SIGN_SIZE + 4,
            ulDataLen, ulBufOffset);
  *(cpMsgBuf + ulDataLen) = _INT32;
  memcpy(cpMsgBuf + ulDataLen + _BL_SIGN_SIZE, (char *) &(i32Number.value), 4);
  ulDataLen += (_BL_SIGN_SIZE + 4);
  return *this;

}


BL_Message&
BL_Message::operator<<(BL_Uint32 ui32Number)
{

  //
  // BL_Message method to put an BL_Uint32 integer in the message.
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: ResizeMsg;
  //
  // Parameters:   BL_Uint32 ui32Number        BL_Uint32 value
  //
  // Returns:      the modified BL_Message
  //


  ResizeMsg(cpMsgBuf, ulDataLen + _BL_SIGN_SIZE + 4,
            ulDataLen, ulBufOffset);
  *(cpMsgBuf + ulDataLen) = _UINT32;
  memcpy(cpMsgBuf + ulDataLen + _BL_SIGN_SIZE, (char *) &(ui32Number.value), 4);
  ulDataLen += (_BL_SIGN_SIZE + 4);
  return *this;

}


BL_Message&
BL_Message::operator<<(BL_Int64 i64Number)
{

  //
  // BL_Message method to put an BL_Int64 integer in the message.
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: ResizeMsg;
  //
  // Parameters:   BL_Int64 i64Number          BL_Int64 value
  //
  // Returns:      the modified BL_Message
  //


  ResizeMsg(cpMsgBuf, ulDataLen + _BL_SIGN_SIZE + 8,
            ulDataLen, ulBufOffset);
  *(cpMsgBuf + ulDataLen) = _INT64;
  memcpy(cpMsgBuf + ulDataLen + _BL_SIGN_SIZE, (char *) &(i64Number.value), 8);
  ulDataLen += (_BL_SIGN_SIZE + 8);
  return *this;

}


BL_Message&
BL_Message::operator<<(BL_Uint64 ui64Number)
{

  //
  // BL_Message method to put an BL_Uint64 integer in the message.
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: ResizeMsg;
  //
  // Parameters:   BL_Uint64 ui64Number        BL_Uint64 value
  //
  // Returns:      the modified BL_Message
  //


  ResizeMsg(cpMsgBuf, ulDataLen + _BL_SIGN_SIZE + 8,
            ulDataLen, ulBufOffset);
  *(cpMsgBuf + ulDataLen) = _UINT64;
  memcpy(cpMsgBuf + ulDataLen + _BL_SIGN_SIZE, (char *) &(ui64Number.value), 8);
  ulDataLen += (_BL_SIGN_SIZE + 8);
  return *this;

}


BL_Message&
BL_Message::operator<<(BL_Charp cpString)
{

  //
  // BL_Message method to put a string in the message.
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: ResizeMsg;
  //
  // Parameters:   BL_Charp cpString           null terminated string
  //
  // Returns:      the modified BL_Message
  //


  unsigned long ulStrLength;
  unsigned char ucaLength[4];
  int iLenSize;

  // Calculates string length & string length dimension
  ulStrLength = strlen(cpString);
  iLenSize = CalcLen(ucaLength, ulStrLength);

  // Reallocates message buffer
  ResizeMsg(cpMsgBuf, ulDataLen + _BL_SIGN_SIZE + iLenSize + ulStrLength,
            ulDataLen, ulBufOffset);

  // Stores type signature (warning: short cut for value!)
  *(cpMsgBuf + ulDataLen) = _CHARPT + iLenSize - 1;

  // Stores string length
  memcpy(cpMsgBuf + ulDataLen + _BL_SIGN_SIZE, ucaLength, iLenSize);

  // Stores string without null character
  if (ulStrLength)
    memcpy(cpMsgBuf + ulDataLen + _BL_SIGN_SIZE + iLenSize,
           cpString, ulStrLength);

  ulDataLen += (_BL_SIGN_SIZE + iLenSize + ulStrLength);
  BL_convResult = BL_CONV_OK;
  return *this;

}


BL_Message&
BL_Message::operator<<(BL_RawData BinData)
{

  //
  // BL_Message method to put a binary object in the message "as is".
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: ResizeMsg;
  //
  // Parameters:   BL_RawData BinData           BL_RawData structure
  //
  // Returns:      the modified BL_Message
  //


  unsigned char ucaLength[4];
  int iLenSize;

  // Calculates binary data length dimension
  iLenSize = CalcLen(ucaLength, BinData.ulDataLen);

  // Reallocates message buffer
  ResizeMsg(cpMsgBuf, ulDataLen + _BL_SIGN_SIZE + iLenSize + BinData.ulDataLen,
            ulDataLen, ulBufOffset);

  // Stores type signature (warning: short cut for value!)
  *(cpMsgBuf + ulDataLen) = _RDATAT + iLenSize - 1;

  // Stores binary data length
  memcpy(cpMsgBuf + ulDataLen + _BL_SIGN_SIZE, ucaLength, iLenSize);

  // Stores proper binary data
  if (BinData.ulDataLen)
    memcpy(cpMsgBuf + ulDataLen + _BL_SIGN_SIZE + iLenSize,
           BinData.pcData, BinData.ulDataLen);

  ulDataLen += (_BL_SIGN_SIZE + iLenSize + BinData.ulDataLen);
  BL_convResult = BL_CONV_OK;
  return *this;

}


//******************************************************************
//********************* EXTRACTION OPERATORS ***********************
//******************************************************************

// Here follows the family of overloaded extraction operators on class
// BL_Message.
// Each operator works on the object this way:
// 1) extracts the next item in the message (an item is made up of a
//    type, a dimension [if necessary] and formatted data);
// 2) checks if the type of the variable passed matches the type of
//    the item;
// 3) if the type does not match, checks if it is possible to fit the
//    item in the variable (only float/double, char/short/int/long);
// 4) returns the value of the item in the variable or a null value if
//    it was not possible to degrade it; the public variable BL_convResult
//    is set to a value that represents the result of the operation;
//    upon succesful operation, ulBufOffset is moved to the next item
//    available, if any. If BL_OCK_MISMATCH is not defined, the item
//    pointer is updated in any case.
//    If there are no items remaining, ulBufOffset is set to ulDataLen.
//    ulCurrentItem is updated accordingly.
// All numeric values are always stored in the message in the Big Endian
// format: they are re-converted to the memory architecture of the
// machine on extraction, if necessary.
// A special care has been taken to account degrading operations on
// integer (e.g. BL_Int32 to BL_Int16).
// Hence, at the reception, if the size of the destination var is smaller
// than the original size, the conversion function works this way:
// - on signed integers, checks the sign and verifies that the upper bytes
//   of the original value (those that are lost in the conversion) are all
//   zero (positive) or 0xff (negative); if this is true, conversion takes
//   place on the following conditions:
//   1) source is negative and most significant bit of most significant
//      byte remaining is 1;
//   2) source is positive and most significant bit of most significant
//      byte remaining is 0;
// - on unsigned integers, it is tested just the null-state of lost bytes.
//
// The upgrading int -> fp or the degrading fp -> int, as the degrading
// double ->float, are signaled through a positive value for BL_convResult.
//
// Values for BL_convResult:
//
// Warnings   BL_PREC_LOSS       Precision loss (double->float)
//            BL_INT_TO_FP       Integer approxximated to floating point
//            BL_FP_TO_INT       Floating point approxximated in integer
//
// Correct    BL_CONV_OK         Conversion OK
//
// Errors     BL_NO_ITEM         Item pointer beyond last item
//            BL_MISMATCH        Type mismatch
//            BL_CANT_SHRINK     Data value could not fit in
//                               the variable passed
//            BL_NEG_TO_POS      Negative value could not be stored
//                               in the unsigned variable passed


BL_Message&
BL_Message::operator>>(char &cValue)
{
  //
  // BL_Message method to retrieve a signed character from the message.
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: BL_Message::SkipItem;
  //               GiveSizeOf(int);
  //               SignatureToPreset;
  //               FromIntInInt;
  //               FromUintInInt;
  //               RealToInt;
  //
  // Parameters:   char &cValue             character value
  //
  // Returns:      the modified BL_Message
  //
  // Note:         cValue overwritten only if conversion succesfull
  //


  char cType, cNumber;
  int iResult = BL_MISMATCH;


  if (ulBufOffset == ulDataLen)
  {
    BL_convResult = BL_NO_ITEM;
    return *this;
  }

  cType = SignatureToPreset(*(cpMsgBuf + ulBufOffset));

  if ((cType >= BL_CHAR) && (cType <= BL_UI64))
    if (cType < BL_UCHAR) // Signed int stored in char
      iResult = FromIntInInt(cpMsgBuf + ulBufOffset + _BL_SIGN_SIZE,
                             (int) GiveSizeOf(cType), cNumber);

    else // Unsigned int stored in char
      iResult = FromUintInInt(cpMsgBuf + ulBufOffset + _BL_SIGN_SIZE,
                              (int) GiveSizeOf(cType), cNumber);

  #ifdef BL_FP_IEEE_754

   if ((cType == BL_FLOAT) || (cType == BL_DOUBLE))
     iResult = RealToInt(cpMsgBuf + ulBufOffset + _BL_SIGN_SIZE,
                         cType, cNumber, _BL_CHAR);

  #endif

  #ifdef BL_OCK_MISMATCH

   if (iResult >= BL_CONV_OK)
   {
     cValue = cNumber;
     this->SkipItem(1);
   }

  #else

   if (iResult >= BL_CONV_OK)
     cValue = cNumber;
   this->SkipItem(1);

  #endif

  BL_convResult = iResult;
  return *this;
}


BL_Message&
BL_Message::operator>>(unsigned char &ucValue)
{
  //
  // BL_Message method to retrieve an unsigned character from the message.
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: BL_Message::SkipItem;
  //               GiveSizeOf(int);
  //               SignatureToPreset;
  //               FromIntInUint;
  //               FromUintInUint;
  //               RealToInt;
  //
  // Parameters:   uchar &ucValue           character value
  //
  // Returns:      the modified BL_Message
  //
  // Note:         ucValue overwritten only if conversion
  //               succesfull
  //


  char cType;
  unsigned char ucNumber;
  int iResult = BL_MISMATCH;


  if (ulBufOffset == ulDataLen)
  {
    BL_convResult = BL_NO_ITEM;
    return *this;
  }

  cType = SignatureToPreset(*(cpMsgBuf + ulBufOffset));

  if ((cType >= BL_CHAR) && (cType <= BL_UI64))
    if (cType >= BL_UCHAR) // Unsigned int stored in unsigned char
      iResult = FromUintInUint(cpMsgBuf + ulBufOffset + _BL_SIGN_SIZE,
                               (int) GiveSizeOf(cType), ucNumber);

    else // Signed int stored in unsigned char
      iResult = FromIntInUint(cpMsgBuf + ulBufOffset + _BL_SIGN_SIZE,
                              (int) GiveSizeOf(cType), ucNumber);

  #ifdef BL_FP_IEEE_754

   if ((cType == BL_FLOAT) || (cType == BL_DOUBLE))
     iResult = RealToInt(cpMsgBuf + ulBufOffset + _BL_SIGN_SIZE,
                         cType, ucNumber, _BL_UCHAR);

  #endif

  #ifdef BL_OCK_MISMATCH

   if (iResult >= BL_CONV_OK)
   {
     ucValue = ucNumber;
     this->SkipItem(1);
   }

  #else

   if (iResult >= BL_CONV_OK)
     ucValue = ucNumber;
   this->SkipItem(1);

  #endif

  BL_convResult = iResult;
  return *this;
}


BL_Message&
BL_Message::operator>>(short &sValue)
{
  //
  // BL_Message method to retrieve a signed short from the message.
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: BL_Message::SkipItem;
  //               GiveSizeOf(int);
  //               SignatureToPreset;
  //               FromIntInInt;
  //               FromUintInInt;
  //               RealToInt;
  //
  // Parameters:   short &sValue            short value
  //
  // Returns:      the modified BL_Message
  //
  // Note:         sValue overwritten only if conversion
  //               succesfull
  //


  char cType;
  short sNumber;
  int iResult = BL_MISMATCH;


  if (ulBufOffset == ulDataLen)
  {
    BL_convResult = BL_NO_ITEM;
    return *this;
  }

  cType = SignatureToPreset(*(cpMsgBuf + ulBufOffset));

  if ((cType >= BL_CHAR) && (cType <= BL_UI64))
    if (cType < BL_UCHAR) // Signed int stored in short
      iResult = FromIntInInt(cpMsgBuf + ulBufOffset + _BL_SIGN_SIZE,
                             (int) GiveSizeOf(cType), sNumber);

    else // Unsigned int stored in short
      iResult = FromUintInInt(cpMsgBuf + ulBufOffset + _BL_SIGN_SIZE,
                              (int) GiveSizeOf(cType), sNumber);

  #ifdef BL_FP_IEEE_754

   if ((cType == BL_FLOAT) || (cType == BL_DOUBLE))
     iResult = RealToInt(cpMsgBuf + ulBufOffset + _BL_SIGN_SIZE,
                         cType, sNumber, _BL_SHORT);

  #endif

  #ifdef BL_OCK_MISMATCH

   if (iResult >= BL_CONV_OK)
   {
     sValue = sNumber;
     this->SkipItem(1);
   }

  #else

   if (iResult >= BL_CONV_OK)
     sValue = sNumber;
   this->SkipItem(1);

  #endif

  BL_convResult = iResult;
  return *this;
}


BL_Message&
BL_Message::operator>>(unsigned short &usValue)
{
  //
  // BL_Message method to retrieve an unsigned short from the message.
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: BL_Message::SkipItem;
  //               GiveSizeOf(int);
  //               SignatureToPreset;
  //               FromIntInInt;
  //               FromIntInUint;
  //               RealToInt;
  //
  // Parameters:   ushort &usValue          unsigned short value
  //
  // Returns:      the modified BL_Message
  //
  // Note:         usValue overwritten only if conversion
  //               succesfull
  //


  char cType;
  unsigned short usNumber;
  int iResult = BL_MISMATCH;


  if (ulBufOffset == ulDataLen)
  {
    BL_convResult = BL_NO_ITEM;
    return *this;
  }

  cType = SignatureToPreset(*(cpMsgBuf + ulBufOffset));

  if ((cType >= BL_CHAR) && (cType <= BL_UI64))
    if (cType >= BL_UCHAR) // Unsigned int stored in unsigned short
      iResult = FromUintInUint(cpMsgBuf + ulBufOffset + _BL_SIGN_SIZE,
                               (int) GiveSizeOf(cType), usNumber);

    else // Signed int stored in unsigned short
      iResult = FromIntInUint(cpMsgBuf + ulBufOffset + _BL_SIGN_SIZE,
                              (int) GiveSizeOf(cType), usNumber);

  #ifdef BL_FP_IEEE_754

   if ((cType == BL_FLOAT) || (cType == BL_DOUBLE))
     iResult = RealToInt(cpMsgBuf + ulBufOffset + _BL_SIGN_SIZE,
                         cType, usNumber, _BL_USHORT);

  #endif

  #ifdef BL_OCK_MISMATCH

   if (iResult >= BL_CONV_OK)
   {
     usValue = usNumber;
     this->SkipItem(1);
   }

  #else

   if (iResult >= BL_CONV_OK)
     usValue = usNumber;
   this->SkipItem(1);

  #endif

  BL_convResult = iResult;
  return *this;
}


BL_Message&
BL_Message::operator>>(int &iValue)
{
  //
  // BL_Message method to retrieve a signed int from the message.
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: BL_Message::SkipItem;
  //               GiveSizeOf(int);
  //               SignatureToPreset;
  //               FromIntInInt;
  //               FromUintInInt;
  //               RealToInt;
  //
  // Parameters:   int &iValue              int value
  //
  // Returns:      the modified BL_Message
  //
  // Note:         iValue overwritten only if conversion
  //               succesfull
  //


  char cType;
  int iNumber;
  int iResult = BL_MISMATCH;


  if (ulBufOffset == ulDataLen)
  {
    BL_convResult = BL_NO_ITEM;
    return *this;
  }

  cType = SignatureToPreset(*(cpMsgBuf + ulBufOffset));

  if ((cType >= BL_CHAR) && (cType <= BL_UI64))
    if (cType < BL_UCHAR) // Signed int stored in int
      iResult = FromIntInInt(cpMsgBuf + ulBufOffset + _BL_SIGN_SIZE,
                             (int) GiveSizeOf(cType), iNumber);

    else // Unsigned int stored in int
      iResult = FromUintInInt(cpMsgBuf + ulBufOffset + _BL_SIGN_SIZE,
                              (int) GiveSizeOf(cType), iNumber);

  #ifdef BL_FP_IEEE_754

   if ((cType == BL_FLOAT) || (cType == BL_DOUBLE))
     iResult = RealToInt(cpMsgBuf + ulBufOffset + _BL_SIGN_SIZE,
                         cType, iNumber, _BL_INT);

  #endif

  #ifdef BL_OCK_MISMATCH

   if (iResult >= BL_CONV_OK)
   {
     iValue = iNumber;
     this->SkipItem(1);
   }

  #else

   if (iResult >= BL_CONV_OK)
     iValue = iNumber;
   this->SkipItem(1);

  #endif

  BL_convResult = iResult;
  return *this;
}


BL_Message&
BL_Message::operator>>(unsigned int &uiValue)
{
  //
  // BL_Message method to retrieve an unsigned int from the message.
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: BL_Message::SkipItem;
  //               GiveSizeOf(int);
  //               SignatureToPreset;
  //               FromIntInUint;
  //               FromUintInUint;
  //               RealToInt;
  //
  // Parameters:   uint &uiValue            unsigned int value
  //
  // Returns:      the modified BL_Message
  //
  // Note:         uiValue overwritten only if conversion
  //               succesfull
  //


  char cType;
  unsigned int uiNumber;
  int iResult = BL_MISMATCH;


  if (ulBufOffset == ulDataLen)
  {
    BL_convResult = BL_NO_ITEM;
    return *this;
  }

  cType = SignatureToPreset(*(cpMsgBuf + ulBufOffset));

  if ((cType >= BL_CHAR) && (cType <= BL_UI64))
    if (cType >= BL_UCHAR) // Unsigned int stored in unsigned int
      iResult = FromUintInUint(cpMsgBuf + ulBufOffset + _BL_SIGN_SIZE,
                               (int) GiveSizeOf(cType), uiNumber);

    else // Signed int stored in unsigned int
      iResult = FromIntInUint(cpMsgBuf + ulBufOffset + _BL_SIGN_SIZE,
                              (int) GiveSizeOf(cType), uiNumber);

  #ifdef BL_FP_IEEE_754

   if ((cType == BL_FLOAT) || (cType == BL_DOUBLE))
     iResult = RealToInt(cpMsgBuf + ulBufOffset + _BL_SIGN_SIZE,
                         cType, uiNumber, _BL_UINT);

  #endif

  #ifdef BL_OCK_MISMATCH

   if (iResult >= BL_CONV_OK)
   {
     uiValue = uiNumber;
     this->SkipItem(1);
   }

  #else

   if (iResult >= BL_CONV_OK)
     uiValue = uiNumber;
   this->SkipItem(1);

  #endif

  BL_convResult = iResult;
  return *this;
}


BL_Message&
BL_Message::operator>>(long &lValue)
{
  //
  // BL_Message method to retrieve a signed long from the message.
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: BL_Message::SkipItem;
  //               GiveSizeOf(int);
  //               SignatureToPreset;
  //               FromIntInInt;
  //               FromUintInInt;
  //               RealToInt;
  //
  // Parameters:   long &lValue             long value
  //
  // Returns:      the modified BL_Message
  //
  // Note:         lValue overwritten only if conversion
  //               succesfull
  //


  char cType;
  long lNumber;
  int iResult = BL_MISMATCH;


  if (ulBufOffset == ulDataLen)
  {
    BL_convResult = BL_NO_ITEM;
    return *this;
  }

  cType = SignatureToPreset(*(cpMsgBuf + ulBufOffset));

  if ((cType >= BL_CHAR) && (cType <= BL_UI64))
    if (cType < BL_UCHAR) // Signed int stored in long
      iResult = FromIntInInt(cpMsgBuf + ulBufOffset + _BL_SIGN_SIZE,
                             (int) GiveSizeOf(cType), lNumber);

    else // Unsigned int stored in long
      iResult = FromUintInInt(cpMsgBuf + ulBufOffset + _BL_SIGN_SIZE,
                               (int) GiveSizeOf(cType), lNumber);

  #ifdef BL_FP_IEEE_754

   if ((cType == BL_FLOAT) || (cType == BL_DOUBLE))
     iResult = RealToInt(cpMsgBuf + ulBufOffset + _BL_SIGN_SIZE,
                         cType, lNumber, _BL_LONG);

  #endif

  #ifdef BL_OCK_MISMATCH

   if (iResult >= BL_CONV_OK)
   {
     lValue = lNumber;
     this->SkipItem(1);
   }

  #else

   if (iResult >= BL_CONV_OK)
     lValue = lNumber;
   this->SkipItem(1);

  #endif

  BL_convResult = iResult;
  return *this;
}


BL_Message&
BL_Message::operator>>(unsigned long &ulValue)
{
  //
  // BL_Message method to retrieve an unsigned long from the message.
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: BL_Message::SkipItem;
  //               GiveSizeOf(int);
  //               SignatureToPreset;
  //               FromIntInUint;
  //               FromUintInUint;
  //               RealToInt;
  //
  // Parameters:   ulong &ulValue            unsigned long value
  //
  // Returns:      the modified BL_Message
  //
  // Note:         ulValue overwritten only if conversion
  //               succesfull
  //


  char cType;
  unsigned long ulNumber;
  int iResult = BL_MISMATCH;


  if (ulBufOffset == ulDataLen)
  {
    BL_convResult = BL_NO_ITEM;
    return *this;
  }

  cType = SignatureToPreset(*(cpMsgBuf + ulBufOffset));

  if ((cType >= BL_CHAR) && (cType <= BL_UI64))
    if (cType >= BL_UCHAR) // Unsigned int stored in unsigned long
      iResult = FromUintInUint(cpMsgBuf + ulBufOffset + _BL_SIGN_SIZE,
                               (int) GiveSizeOf(cType), ulNumber);

    else // Signed int stored in unsigned long
      iResult = FromIntInUint(cpMsgBuf + ulBufOffset + _BL_SIGN_SIZE,
                              (int) GiveSizeOf(cType), ulNumber);

  #ifdef BL_FP_IEEE_754

   if ((cType == BL_FLOAT) || (cType == BL_DOUBLE))
     iResult = RealToInt(cpMsgBuf + ulBufOffset + _BL_SIGN_SIZE,
                         cType, ulNumber, _BL_ULONG);

  #endif

  #ifdef BL_OCK_MISMATCH

   if (iResult >= BL_CONV_OK)
   {
     ulValue = ulNumber;
     this->SkipItem(1);
   }

  #else

   if (iResult >= BL_CONV_OK)
     ulValue = ulNumber;
   this->SkipItem(1);

  #endif

  BL_convResult = iResult;
  return *this;
}


#ifdef BL_FP_IEEE_754

 BL_Message&
 BL_Message::operator>>(float &fValue)
 {
   //
   // BL_Message method to retrieve a float from the message.
   //
   // Author:       Simone Zanella
   //
   // Version:      1.1
   //
   // Dependencies: BL_Message::SkipItem;
   //               SignatureToPreset;
   //               IntToReal;
   //
   // Parameters:   float &fValue            float value
   //
   // Returns:      the modified BL_Message
   //
   // Note:         fValue overwritten only if conversion
   //               succesfull
   //


   char cType;
   float fNumber;
   int iResult = BL_MISMATCH;


   if (ulBufOffset == ulDataLen)
   {
     BL_convResult = BL_NO_ITEM;
     return *this;
   }

   cType = SignatureToPreset(*(cpMsgBuf + ulBufOffset));
   if ((cType == BL_FLOAT) || (cType == BL_DOUBLE))
     if (cType == BL_FLOAT) // Float stored in a float
     {

       PutReal(cpMsgBuf + ulBufOffset + _BL_SIGN_SIZE, (char *) &fNumber,
               _BL_FLOAT_SIZE);
       iResult = BL_CONV_OK;

     }

     else // Double stored in a float
       iResult = DoubleToFloat(cpMsgBuf + ulBufOffset + _BL_SIGN_SIZE,
                               fNumber);
   else
     if ((cType >= BL_CHAR) && (cType <= BL_UI64))
       iResult = IntToReal(*this, cType, fNumber);

   #ifdef BL_OCK_MISMATCH

    if (iResult >= BL_CONV_OK)
    {
      fValue = fNumber;
      this->SkipItem(1);
    }

   #else

    if (iResult >= BL_CONV_OK)
      fValue = fNumber;
    this->SkipItem(1);

   #endif

   BL_convResult = iResult;
   return *this;
 }


 BL_Message&
 BL_Message::operator>>(double &dValue)
 {
   //
   // BL_Message method to retrieve a double from the message.
   //
   // Author:       Simone Zanella
   //
   // Version:      1.1
   //
   // Dependencies: BL_Message::SkipItem;
   //               SignatureToPreset;
   //               IntToReal;
   //
   // Parameters:   double &dValue           double value
   //
   // Returns:      the modified BL_Message
   //
   // Note:         dValue overwritten only if conversion
   //               succesfull
   //


   char cType;
   float fNumber;
   int iResult = BL_MISMATCH;


   if (ulBufOffset == ulDataLen)
   {
     BL_convResult = BL_NO_ITEM;
     return *this;
   }

   cType = SignatureToPreset(*(cpMsgBuf + ulBufOffset));
   if ((cType == BL_FLOAT) || (cType == BL_DOUBLE))
   {
     iResult = BL_CONV_OK;

     if (cType == BL_FLOAT) // Float stored in a double
     {

       PutReal(cpMsgBuf + ulBufOffset + _BL_SIGN_SIZE,
               (char *) &fNumber, _BL_FLOAT_SIZE);
       dValue = (double) fNumber;

     }

     else // Double stored in a double
       PutReal(cpMsgBuf + ulBufOffset + _BL_SIGN_SIZE,
               (char *) &dValue, _BL_DOUBLE_SIZE);
   }
   else
     if ((cType >= BL_CHAR) && (cType <= BL_UI64))
       iResult = IntToReal(*this, cType, dValue);

   #ifdef BL_OCK_MISMATCH

    if (iResult >= BL_CONV_OK)
      this->SkipItem(1);

   #else

    this->SkipItem(1);

   #endif

   BL_convResult = iResult;
   return *this;
 }

#endif //BL_FP_IEEE_754


BL_Message&
BL_Message::operator>>(BL_Charp &cpValue)
{
  //
  // BL_Message method to retrieve a null terminated string.
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: BL_Message::SkipItem;
  //               SignatureToPreset;
  //               BL_ChkNul;
  //               GiveSizeOf(char *);
  //
  // Parameters:   BL_Charp &lValue            string pointer
  //
  // Returns:      the modified BL_Message
  //
  // Note:         this function dinamically allocate memory as
  //               needed; NEVER allocate memory to the variable
  //               before using this operator, because you would
  //               lose the opportunity to release the memory
  //               previously reserved.
  //


  char cType;
  unsigned long ulSize;
  int iResult = BL_MISMATCH;


  if (ulBufOffset == ulDataLen)
  {
    BL_convResult = BL_NO_ITEM;
    return *this;
  }

  cType = SignatureToPreset(*(cpMsgBuf + ulBufOffset));

  if (cType == BL_CHARP)
  {
    // Reserves space for string + null character

    ulSize = GiveSizeOf(cpMsgBuf + ulBufOffset) + 1;
    BL_ChkNul(cpValue = (char *) malloc(ulSize));

    if (ulSize>1) // Copies string, if it is not null
      memcpy(cpValue, cpMsgBuf + ulBufOffset + _BL_SIGN_SIZE +
             *(cpMsgBuf + ulBufOffset) - _CHARPT + 1, ulSize - 1);

    // Adds null character
    *(cpValue + ulSize - 1) = 0;

    // Updates item's pointer
    this->SkipItem(1);
    iResult = BL_CONV_OK;
  }

  #ifndef BL_OCK_MISMATCH

   if (iResult < BL_CONV_OK)
     this->SkipItem(1);

  #endif

  BL_convResult = iResult;
  return *this;

}


BL_Message&
BL_Message::operator>>(BL_RawData &rdValue)
{
  //
  // BL_Message method to retrieve a binary object (BL_RawData).
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: BL_Message::SkipItem;
  //               SignatureToPreset;
  //               BL_ChkNul;
  //               GiveSizeOf(char *);
  //
  // Parameters:   BL_RawData &rdValue         BL_RawData structure
  //
  // Returns:      the modified BL_Message
  //
  // Note:         this function dinamically allocate memory as
  //               needed; NEVER allocate memory to the variable
  //               before using this operator, because you would
  //               lose the opportunity to release the memory
  //               previously reserved.
  //


  char cType;
  unsigned long ulSize;
  int iResult = BL_MISMATCH;


  if (ulBufOffset == ulDataLen)
  {
    BL_convResult = BL_NO_ITEM;
    return *this;
  }

  cType = SignatureToPreset(*(cpMsgBuf + ulBufOffset));

  if (cType == BL_RAWDATA)
  {
    ulSize = GiveSizeOf(cpMsgBuf + ulBufOffset);
    rdValue.ulDataLen = ulSize;

    if (ulSize) // Copies the binary object, if its length is not null
    {
      BL_ChkNul(rdValue.pcData = (char *) malloc(ulSize));
      memcpy(rdValue.pcData, cpMsgBuf + ulBufOffset + _BL_SIGN_SIZE +
             *(cpMsgBuf + ulBufOffset) - _RDATAT + 1, ulSize);
    }
    else
      rdValue.pcData = NULL;

    this->SkipItem(1);
    iResult = BL_CONV_OK;
  }

  #ifndef BL_OCK_MISMATCH

   if (iResult < BL_CONV_OK)
     this->SkipItem(1);

  #endif

  BL_convResult = iResult;
  return *this;

}


//******************************************************************
//*********************** LOAD/SAVE METHODS ************************
//******************************************************************

#define SEEK_CUR    1
#define SEEK_END    2
#define SEEK_SET    0
#define BLOCK_SIZE  2048ul

#include <fcntl.h>

#ifdef BL_MSD_INCLUDE
  #include <io.h>
  #include <sys\stat.h>
#else
  #include <sys/stat.h>
  #include <unistd.h>
#endif

int
BL_Message::MsgLoad (char * cpSourceFile)
{
  //
  // BL_Message method to load a message from a file
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: FromUintInUint;
  //               FromIntInInt;
  //               BL_ChkNul;
  //
  // Parameters:   char * cpSourceFile         name of file to read message from
  //
  // Returns:      0                           if error
  //               1                           if successfull
  //


  int hFile;       // File handle
  char caType[4];  // Space to store message type tag (32 bit integer)

  #ifdef BL_MSD_INCLUDE

    if ((hFile = open(cpSourceFile, O_RDONLY | O_BINARY)) == -1)
      return 0;

  #else

    if ((hFile = open(cpSourceFile, O_RDONLY)) == -1)
      return 0;

  #endif

  // Calculates message length, which is file length minus the 4 bytes
  // used for type tag.

  ulDataLen = lseek(hFile, 0L, SEEK_END) - 4;

  // Repositions the file pointer to beginning of file.

  lseek(hFile, 0L, SEEK_SET);

  // Reads type tag information

  if (read(hFile, caType, 4) == -1)
  {
    close(hFile);
    return 0;
  }

  // Re-convert the binary values into machine-processable values.

  FromIntInInt(caType, 4, i32Type);

  // Re-allocates message buffer to store the entire message.

  if (ulDataLen)
    BL_ChkNul(cpMsgBuf = (char *) realloc(cpMsgBuf, ulDataLen));
  else
    cpMsgBuf = NULL;

  // Reads the message from file.

  if (read(hFile, cpMsgBuf, ulDataLen) == -1)
  {
    close(hFile);
    return 0;
  }

  // Closes the file.

  close(hFile);

  return 1;
}


int
BL_Message::MsgSave (char * cpDestName)
{
  //
  // BL_Message method to save a message to a file
  //
  // Author:       Simone Zanella
  //
  // Version:      1.1
  //
  // Dependencies: ToUint;
  //               ToInt;
  //
  // Parameters:   cpDestName                  Destination file name
  //
  // Returns:      0                           if error
  //               1                           if successfull
  //


  int hFile;       // File handle
  char caType[4];  // Space to store message type tag (32 bit integer)
  unsigned long ulTotalByte; // number of bytes to write

  // Opens file cpDestName for writing, creating it if it does not
  // exists or truncating it to zero length if it is already on disk.

  #ifdef BL_MSD_INCLUDE

    if ((hFile = open(cpDestName, O_BINARY | O_CREAT | O_WRONLY | O_TRUNC,
                      S_IWRITE)) == -1)
      return 0;

  #else

    umask(0);
    if ((hFile = open(cpDestName, O_CREAT | O_WRONLY | O_TRUNC | O_APPEND,
                      S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH |
                      S_IWOTH)) == -1)
      return 0;

  #endif

  // Stores message type tag into the 4 byte array;
  // the values are stored in the BLlib memory format
  // (big endian).

  ToInt(caType, 4, i32Type);

  // Writes type information.

  if (write(hFile, caType, 4) == -1)
  {
    close(hFile);
    return 0;
  }

  // Writes message buffer as is.

  ulTotalByte = ulDataLen;
  while (ulTotalByte)
  {
    if (write(hFile, cpMsgBuf + ulDataLen - ulTotalByte,
              BL_min(BLOCK_SIZE, ulTotalByte)) == -1)
    {
      close(hFile);
      return 0;
    }
    ulTotalByte -= BL_min(BLOCK_SIZE, ulTotalByte);
  }
  // Closes the file.

  close(hFile);
  return 1;
}


//******************************************************************
//************************** END OF FILE ***************************
//******************************************************************
